{+--------------------------------------------------------------------------+
 |                 68HC11 Assembler Language Syntax Parser v0.51            |
 +--------------------------------------------------------------------------+
 |
 | Description:   An 68HC11 Assembly Language syntax parser for use
 |                with mwCustomEdit. Supports all 68HC11 Op codes.
 |
 +--------------------------------------------------------------------------+
 | Unit:          CIHC11Syn
 | Created:       2000-01
 | Last change:   2000-01-30
 | Author:        Nils Springob (delphi.nils@crazy-idea.de)
 | Copyright      (c) 2000 All rights reserved
 |                Portions Copyright Martin Waldenburg
 | Version:       0.51
 | Status:        Public Domain
 | DISCLAIMER:    This is provided as is, expressly without a warranty of any kind.
 |                You use it at your own risc.
 |
 | Thanks to:     Martin Waldenburg, Dave Muir, Hideo Koiso, Nick Hoddinott
 +--------------------------------------------------------------------------+
 | See file "Version.rtf" for full Revision History
 +--------------------------------------------------------------------------+}

unit CIHC11Syn;

{$I MWEDIT.INC}

interface

uses
  SysUtils, Windows, Messages, Classes, Controls, Graphics, Registry,
  mwHighlighter;

const
  KeyWordCount = 149;
  DirectiveCount = 6;

type
  TtkTokenKind = (
    tkComment,
    tkDirective,
    tkIdentifier,
    tkKey,
    tkNull,
    tkNumber,
    tkSpace,
    tkString,
    tkSymbol,
    tkUnknown);

  TkwKeyWordType = (
    kwNone,
    kwOperand,
    kwOperandOver,
    kwNoOperand);

  PHashListEntry = ^THashListEntry;
  THashListEntry = record
    Next: PHashListEntry;
    Token: String;
    Kind: TtkTokenKind;
    Op: Boolean;
  end;

  TProcTableProc = procedure of Object;

  TCIHC11Syn = class(TmwCustomHighLighter)
  private
    fLine: PChar;
    fLineNumber: integer;
    fProcTable: array[#0..#255] of TProcTableProc;
    Run: LongInt;
    fStringLen: Integer;
    fToIdent: PChar;
    fTokenPos: Integer;
    FTokenID: TtkTokenKind;
    FKeyWordType: TkwKeyWordType;
    fCommentAttri: TmwHighLightAttributes;
    fDirecAttri: TmwHighLightAttributes;
    fIdentifierAttri: TmwHighLightAttributes;
    fInvalidAttri: TmwHighLightAttributes;
    fKeyAttri: TmwHighLightAttributes;
    fNumberAttri: TmwHighLightAttributes;
    fSpaceAttri: TmwHighLightAttributes;
    fStringAttri: TmwHighLightAttributes;
    fSymbolAttri: TmwHighLightAttributes;

    fHashArray: array[0..DirectiveCount+KeyWordCount] of THashListEntry;
    fHashList: array[#0..#255] of PHashListEntry;
    fHashArrayIndex: integer;

    function KeyHash(ToHash: PChar): char;
    function KeyComp(const aKey: string): Boolean;

    procedure SymAsciiCharProc;
    procedure SymbolProc;
    procedure SymDollarProc;
    procedure SymCRProc;
    procedure SymIdentProc;
    procedure SymLFProc;
    procedure SymPercentProc;
    procedure SymNullProc;
    procedure SymNumberProc;
    procedure SymSpaceProc;
    procedure SymStarProc;
    procedure SymStringProc;
    procedure SymUnknownProc;

    procedure InitIdent;
    procedure MakeMethodTables;
    procedure AddHashEntry(NewToken: String; NewKind: TtkTokenKind);
    function IdentKind(MayBe: PChar): TtkTokenKind;
  protected
    function GetIdentChars: TIdentChars; override;
  public
    class function GetLanguageName: string; override;
    class function GetCapability: THighlighterCapability; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExportNext; override;
    function GetDefaultAttribute(Index: integer): TmwHighLightAttributes;
      override;
    function GetEol: Boolean; override;
    function GetRange: Pointer; override;
    function GetTokenID: TtkTokenKind;
    procedure SetLine(NewValue: String; LineNumber:Integer); override;
    function GetToken: String; override;
    function GetTokenAttribute: TmwHighLightAttributes; override;
    function GetTokenKind: integer; override;
    function GetTokenPos: Integer; override;
    procedure Next; override;
    procedure ReSetRange; override;
    procedure SetRange(Value: Pointer); override;
    procedure SetLineForExport(NewValue: string); override;
  published
    property CommentAttri: TmwHighLightAttributes read fCommentAttri
      write fCommentAttri;
    property DirecAttri: TmwHighLightAttributes read fDirecAttri
      write fDirecAttri;
    property IdentifierAttri: TmwHighLightAttributes read fIdentifierAttri
      write fIdentifierAttri;
    property InvalidAttri: TmwHighLightAttributes read fInvalidAttri
      write fInvalidAttri;
    property KeyAttri: TmwHighLightAttributes read fKeyAttri write fKeyAttri;
    property NumberAttri: TmwHighLightAttributes read fNumberAttri
      write fNumberAttri;
    property SpaceAttri: TmwHighLightAttributes read fSpaceAttri
      write fSpaceAttri;
    property StringAttri: TmwHighLightAttributes read fStringAttri
      write fStringAttri;
    property SymbolAttri: TmwHighLightAttributes read fSymbolAttri
      write fSymbolAttri;
  end;

procedure Register;

implementation

uses
  mwExport, mwLocalStr;

procedure Register;
begin
  RegisterComponents(MWS_HighlightersPage, [TCIHC11Syn]);
end;

var
  Identifiers: array[#0..#255] of ByteBool;

const
  KeyWords: array[1..KeyWordCount] of string = (
    'ABA', 'ABX', 'ABY', 'ADCA_', 'ADCB_', 'ADDA_', 'ADDB_', 'ADDD_', 'ANDA_',
    'ANDB_', 'ASLA', 'ASLB', 'ASL_', 'ASLD', 'ASRA', 'ASRB', 'ASR_', 'BCC_',
    'BCLR_', 'BCS_', 'BEQ_', 'BGE_', 'BGT_', 'BHI_', 'BHS_', 'BITA_', 'BITB_',
    'BLE_', 'BLO_', 'BLS_', 'BLT_', 'BMI_', 'BNE_', 'BPL_', 'BRA_', 'BRCLR_',
    'BRN_', 'BRSET_', 'BSET_', 'BSR_', 'BVC_', 'BVS_', 'CBA', 'CLC', 'CLI',
    'CLRA', 'CLRB', 'CLR_', 'CLV', 'CMPA_', 'CMPB_', 'COMA', 'COMB', 'COM_',
    'CPD_', 'CPX_', 'CPY_', 'DAA', 'DECA', 'DECB', 'DEC_', 'DES', 'DEX', 'DEY',
    'EORA_', 'EORB_', 'FDIV', 'IDIV', 'INCA', 'INCB', 'INC_', 'INS', 'INX',
    'INY', 'JMP_', 'JSR_', 'LDAA_', 'LDAB_', 'LDD_', 'LDS_', 'LDX_', 'LDY_',
    'LSLA', 'LSLB', 'LSL_', 'LSLD', 'LSRA', 'LSRB', 'LSR_', 'LSRD', 'MUL',
    'NEGA', 'NEGB', 'NEG_', 'NOP', 'ORAA_', 'ORAB_', 'PSHA', 'PSHB', 'PSHX',
    'PSHY', 'PULA', 'PULB', 'PULX', 'PULY', 'ROLA', 'ROLB', 'ROL_', 'RORA',
    'RORB', 'ROR_', 'RTI', 'RTS', 'SBA', 'SBCA_', 'SBCB_', 'SEC', 'SEI', 'SEV',
    'STAA_', 'STAB_', 'STD_', 'STOP', 'STS_', 'STX_', 'STY_', 'SUBA_', 'SUBB_',
    'SUBD_', 'SWI', 'TAB', 'TAP', 'TBA', 'TEST', 'TPA', 'TSTA', 'TSTB', 'TST_',
    'TSX', 'TSY', 'TXS', 'TYS', 'WAI', 'XGDX', 'XGDY', // Ende Befehle
    'FCC_','FCB_','BSZ_','FDB_' // Codeerzeugende Direktiven
    );

  Directives: array[1..DirectiveCount] of string = (
    'EQU_', 'OPT_', 'PAGE', 'ORG_', 'RMB_', 'END'  // Direktiven
    );
procedure MakeIdentTable;
var
  I: Char;
begin
  for I := #0 to #255 do
  begin
    Case I of
      '_', '0'..'9', 'a'..'z', 'A'..'Z': Identifiers[I] := True;
    else Identifiers[I] := False;
    end;
  end;
end;


procedure TCIHC11Syn.AddHashEntry(NewToken: String; NewKind: TtkTokenKind);
var
  Hash: char;
  actEntryPtr: ^PHashListEntry;
  actEntry: PHashListEntry;
begin
  with fHashArray[fHashArrayIndex] do
  begin
    Op := false;
    if Pos('_', NewToken) <> 0 then begin
      Delete(NewToken, Length(NewToken), 1);
      Op := true;
    end;
    Token := NewToken;
    Kind := NewKind;
    Hash := KeyHash(PChar(NewToken));
    actEntryPtr := @fHashList[Hash];
    actEntry := fHashList[Hash];
    while (actEntry <> nil) do
    begin
      actEntryPtr := @actEntry.Next;
      actEntry := actEntry.Next;
    end;
    Next := nil;
  end;
  actEntryPtr^ := @fHashArray[fHashArrayIndex];
  inc(fHashArrayIndex);
end;

procedure TCIHC11Syn.InitIdent;
var
  i: integer;
  c: char;
begin
  // initialize list
  fHashArrayIndex := 0;
  for c := #0 to #255 do
    fHashList[c] := nil;
  // put all keywords into the list
  for i := 1 to KeyWordCount do
    AddHashEntry(KeyWords[i], tkKey);
  // put all directives into the list
  for i := 1 to DirectiveCount do
    AddHashEntry(Directives[i], tkDirective);
end;

function TCIHC11Syn.KeyHash(ToHash: PChar): char;
begin
  Result := #0;
  while ToHash^ in ['_', '0'..'9', 'a'..'z', 'A'..'Z'] do
  begin
//    inc (Result, ord(ToHash^));
    Result := char(byte(Result) + byte(ToHash^));                               //eb 2000-01-30
    inc(ToHash);
  end;
  fStringLen := ToHash - fToIdent;
end; { KeyHash }

function TCIHC11Syn.KeyComp(const aKey: String): Boolean;
var
  I: Integer;
  Temp: PChar;
begin
  Temp := fToIdent;
  if Length(aKey) = fStringLen then
  begin
    Result := True;
    for i := 1 to fStringLen do
    begin
      if Temp^ <> aKey[i] then
      begin
        Result := False;
        break;
      end;
      inc(Temp);
    end;
  end else Result := False;
end; { KeyComp }

function TCIHC11Syn.IdentKind(MayBe: PChar): TtkTokenKind;
var
  Hash: char;
  actEntry: PHashListEntry;
begin
  fToIdent := MayBe;
  Hash := KeyHash(MayBe);

  Result := tkIdentifier;

  actEntry := fHashList[Hash];
  while (actEntry <> nil) do
  begin
    if KeyComp(actEntry.Token) then
    begin
      Result := actEntry.Kind;
      if actEntry.Op then
        FKeyWordType := kwOperand
      else
        FKeyWordType := kwNoOperand;
    end;
    actEntry := actEntry.Next;
  end;
end;

procedure TCIHC11Syn.MakeMethodTables;
var
  I: Char;
begin
  for I := #0 to #255 do
    case I of
      #39: fProcTable[I] := SymAsciiCharProc;
      '$': fProcTable[I] := SymDollarProc;
      #13: fProcTable[I] := SymCRProc;
      'A'..'Z', 'a'..'z', '_': fProcTable[I] := SymIdentProc;
      #10: fProcTable[I] := SymLFProc;
      '%': fProcTable[I] := SymPercentProc;
      #0: fProcTable[I] := SymNullProc;
      '0'..'9': fProcTable[I] := SymNumberProc;
      #1..#9, #11, #12, #14..#32: fProcTable[I] := SymSpaceProc;
      '*': fProcTable[I] := SymStarProc;
      #34: fProcTable[I] := SymStringProc;
      '#', ':', ',', ';', '(', ')': fProcTable[I] := SymbolProc;
      else fProcTable[I] := SymUnknownProc;
    end;
end;

constructor TCIHC11Syn.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fCommentAttri := TmwHighLightAttributes.Create(MWS_AttrComment);
  fCommentAttri.Style:= [fsItalic];
  AddAttribute(fCommentAttri);
  fIdentifierAttri := TmwHighLightAttributes.Create(MWS_AttrIdentifier);
  AddAttribute(fIdentifierAttri);
  fInvalidAttri := TmwHighLightAttributes.Create(MWS_AttrIllegalChar);
  AddAttribute(fInvalidAttri);
  fKeyAttri := TmwHighLightAttributes.Create(MWS_AttrReservedWord);
  fKeyAttri.Style:= [fsBold];
  AddAttribute(fKeyAttri);
  fNumberAttri := TmwHighLightAttributes.Create(MWS_AttrNumber);
  AddAttribute(fNumberAttri);
  fSpaceAttri := TmwHighLightAttributes.Create(MWS_AttrSpace);
  AddAttribute(fSpaceAttri);
  fStringAttri := TmwHighLightAttributes.Create(MWS_AttrString);
  AddAttribute(fStringAttri);
  fSymbolAttri := TmwHighLightAttributes.Create(MWS_AttrSymbol);
  AddAttribute(fSymbolAttri);
  fDirecAttri := TmwHighLightAttributes.Create(MWS_AttrPreprocessor);
  AddAttribute(fDirecAttri);
  SetAttributesOnChange(DefHighlightChange);

  MakeMethodTables;
  InitIdent;
  fDefaultFilter := MWS_FilterAsm68HC11;
end; { Create }

procedure TCIHC11Syn.SetLine(NewValue: string; LineNumber: Integer);
begin
  FKeyWordType := kwNone;
  fLine := PChar(NewValue);
  Run := 0;
  fLineNumber := LineNumber;
  Next;
end; { SetLine }

procedure TCIHC11Syn.SymAsciiCharProc;
begin
  fTokenID := tkString;
  if (FLine[Run + 1] = #39) and (FLine[Run + 2] = #39) then inc(Run, 2);
  repeat
    case FLine[Run] of
      #0, #10, #13:
      begin
        FKeyWordType:=kwNone;
        break;
      end;
    end;
    inc(Run);
  until FLine[Run] = #39;
  if FLine[Run] <> #0 then inc(Run);
end;

procedure TCIHC11Syn.SymbolProc;
begin
  fTokenID := tkSymbol;
  inc(Run);
end;

procedure TCIHC11Syn.SymDollarProc;
begin
  fTokenID := tkNumber;
  inc(Run);
  while FLine[Run] in ['0'..'9', 'A'..'F', 'a'..'f'] do
    inc(Run);
end;

procedure TCIHC11Syn.SymCRProc;
begin
  fTokenID := tkSpace;
  FKeyWordType := kwNone;
  inc(Run);
  if fLine[Run] = #10 then inc(Run);
end;

procedure TCIHC11Syn.SymIdentProc;
begin
  fTokenID := IdentKind((fLine + Run));
  inc(Run, fStringLen);
  while Identifiers[fLine[Run]] do inc(Run);
end;

procedure TCIHC11Syn.SymLFProc;
begin
  FKeyWordType := kwNone;
  fTokenID := tkSpace;
  inc(Run);
end;

procedure TCIHC11Syn.SymPercentProc;
begin
  inc(Run);
  fTokenID := tkNumber;
  while FLine[Run] in ['0'..'1'] do
    inc(Run);
end;

procedure TCIHC11Syn.SymNullProc;
begin
  fTokenID := tkNull;
end;

procedure TCIHC11Syn.SymNumberProc;
begin
  inc(Run);
  fTokenID := tkNumber;
  while FLine[Run] in ['0'..'9'] do
    inc(Run);
end;

procedure TCIHC11Syn.SymSpaceProc;
begin
  inc(Run);
  if FKeyWordType in [kwOperandOver, kwNoOperand] then begin
    FKeyWordType := kwNone;
    fTokenID := tkComment;
    while not (fLine[Run] in [#0, #10, #13]) do
      Inc(Run);
  end else begin
    if FKeyWordType = kwOperand then
      FKeyWordType := kwOperandOver;
    fTokenID := tkSpace;
    while (fLine[Run] <= #32) and not (fLine[Run] in [#0, #10, #13]) do
      inc(Run);
  end;
end;

procedure TCIHC11Syn.SymStarProc;
begin
  inc(Run);
  if FKeyWordType = kwOperandOver then
    fTokenID := tkSymbol
  else begin
    fTokenID := tkComment;
    while not (fLine[Run] in [#0, #10, #13]) do
      inc(Run);
  end;
end;

procedure TCIHC11Syn.SymStringProc;
begin
  fTokenID := tkString;
  if (FLine[Run + 1] = #34) and (FLine[Run + 2] = #34) then inc(Run, 2);
  repeat
    case FLine[Run] of
      #0, #10, #13: break;
    end;
    inc(Run);
  until FLine[Run] = #34;
  if FLine[Run] <> #0 then inc(Run);
end;

procedure TCIHC11Syn.SymUnknownProc;
begin
  inc(Run);
  fTokenID := tkUnknown;
end;

procedure TCIHC11Syn.Next;
begin
  fTokenPos := Run;
  fProcTable[fLine[Run]];
end;

function TCIHC11Syn.GetDefaultAttribute(Index: integer): TmwHighLightAttributes;
begin
  case Index of
    MW_DEFATTR_COMMENT: Result := fCommentAttri;
    MW_DEFATTR_KEYWORD: Result := fKeyAttri;
    MW_DEFATTR_WHITESPACE: Result := fSpaceAttri;
    else Result := nil;
  end;
end;

function TCIHC11Syn.GetEol: Boolean;
begin
  Result := (fTokenId = tkNull);
end;

function TCIHC11Syn.GetRange: Pointer;
begin
  Result := nil;
end;

procedure TCIHC11Syn.SetRange(Value: Pointer);
begin
end;

procedure TCIHC11Syn.ReSetRange;
begin
end;

function TCIHC11Syn.GetToken: String;
var
  Len: LongInt;
begin
  Len := Run - fTokenPos;
  SetString(Result, (FLine + fTokenPos), Len);
end;

function TCIHC11Syn.GetTokenAttribute: TmwHighLightAttributes;
begin
  case fTokenID of
    tkComment: Result := fCommentAttri;
    tkDirective: Result := fDirecAttri;
    tkIdentifier: Result := fIdentifierAttri;
    tkKey: Result := fKeyAttri;
    tkNumber: Result := fNumberAttri;
    tkSpace: Result := fSpaceAttri;
    tkString: Result := fStringAttri;
    tkSymbol: Result := fSymbolAttri;
    tkUnknown: Result := fIdentifierAttri;
    else Result := nil;
  end;
end;

function TCIHC11Syn.GetTokenKind: integer;
begin
  Result := Ord(fTokenId);
end;

function TCIHC11Syn.GetTokenID: TtkTokenKind;
begin
  Result := fTokenId;
end;

function TCIHC11Syn.GetTokenPos: Integer;
begin
  Result := fTokenPos;
end;

function TCIHC11Syn.GetIdentChars: TIdentChars;
begin
  Result := ['_', '0'..'9', 'a'..'z', 'A'..'Z'];
end;

class function TCIHC11Syn.GetLanguageName: string;
begin
  Result := MWS_Lang68HC11;
end;

class function TCIHC11Syn.GetCapability: THighlighterCapability;
begin
  Result := inherited GetCapability + [hcExportable];
end;

procedure TCIHC11Syn.SetLineForExport(NewValue: String);
begin
  FKeyWordType := kwNone;
  fLine := PChar(NewValue);
  Run := 0;
  ExportNext;
end;

procedure TCIHC11Syn.ExportNext;
begin
  Next;
  if Assigned(Exporter) then
    with TmwCustomExport(Exporter) do
      case GetTokenID of
        tkComment: FormatToken(GetToken, fCommentAttri, True, False);
        tkDirective: FormatToken(GetToken, fDirecAttri, True, False);
        tkIdentifier: FormatToken(GetToken, fIdentifierAttri, False, False);
        tkKey: FormatToken(GetToken, fKeyAttri, False, False);
        // Needed to catch line breaks
        tkNull: FormatToken('', nil, False, False);
        tkNumber: FormatToken(GetToken, fNumberAttri, False, False);
        tkSpace: FormatToken(GetToken, fSpaceAttri, False, True);
        tkString: FormatToken(GetToken, fStringAttri, True, False);
        tkSymbol: FormatToken(GetToken, fSymbolAttri, True, False);
        tkUnknown: FormatToken(GetToken, fInvalidAttri, False, False);
      end;
end;

initialization
  MakeIdentTable;
  RegisterPlaceableHighlighter(TCIHC11Syn);                                     //gp 2000-01-20
end.
