{$I MWEDIT.INC}

unit mwHighlighterManager;

interface

uses
  Classes;

type
  {:Highlighter manager.<p>
    Design-only component, designed to simplify work with highlighter components.<p>
    When placed on the form, mwHighlighterManager scans the form for highlighter
    components (descendants of TmwHighlighter). Next it presents the user with
    small form containing checkboxed list and some buttons. User can select (by
    checking/unchecking list items) highlighter that should be placed onto the
    form. After user clicks OK, mwHighlighterManager synchronises highlighter
    components on the form with required state.<p>
    Built-in tricks:<br>
    - mwHighlighterManager never covers existing TComponent with a highlighter.<br>
    - mwHighlighterManager scans the form for TmwHighlighter descendants and
      uses topmost and leftmost component as a starting point for insertion. If
      no TmwHighlighter components are found, first highlighter will be placed at
      coordinates (8,8).<p>
    Known issues:<br>
    - If you place TmwHighlighterManager by double-clicking its icon in
      component palette, it will function normally, except that when all is
      done, Delphi will disply small window with title "Error" and message
      "Operation aborted". Purely cosmetic issue for which there is no obvious
      workaround. Live with it.<p>
    Last change: 2000-01-21

    @author Primoz Gabrijelcic
    @version 0.1
    @since mwCustomEdit 0.92
    @component
    @see TmwHighlighter
  :}
  TmwHighlighterManager = class(TComponent)
  public
    constructor Create(AOwner: TComponent); override;
  published
  end;

procedure Register;

implementation

uses
  DsgnIntf,
  Windows,
  SysUtils,
  Forms,
  Controls,
  StdCtrls,
  CheckLst,
  mwLocalStr,
  mwHighlighter;

type
  TmwHighlighterForm = class(TForm)
    clbHighlighters: TCheckListBox;
    btnSelectAll: TButton;
    btnDeselectAll: TButton;
    btnOK: TButton;
    btnCancel: TButton;
    Highlight: THighlighterList;
    constructor Create(highlighters: THighlighterList); {$IFDEF MWE_COMPILER_4_UP}reintroduce;{$ENDIF}
    procedure   LoadForm;
    procedure   SelectAll(Sender: TObject);
    procedure   DeselectAll(Sender: TObject);
  end;

procedure Register;
begin
  RegisterComponents(MWS_ComponentsPage, [TmwHighlighterManager]);
end;

{ TmwHighlighterManager }

constructor TmwHighlighterManager.Create(AOwner: TComponent);
var
  form: TCustomForm;
  dsgn: {$IFDEF MWE_COMPILER_4_UP}IFormDesigner{$ELSE}TFormDesigner{$ENDIF};
  highlight: THighlighterList;
  mwForm: TmwHighlighterForm;

  procedure CheckExisting;
  var
    i: integer;
    j: integer;
  begin
    for i := 0 to form.ComponentCount-1 do begin
      j := highlight.FindByClass(form.Components[i]);
      if j >= 0 then begin
        j := mwForm.clbHighlighters.Items.IndexOf(highlight[j].GetLanguageName);
        if j >= 0 then
          mwForm.clbHighlighters.Checked[j] := true;
      end;
    end; //for
  end;

  function FindHighlighterComp(hlClass: TmwCustomHighlighterClass): integer;
  var
    i: integer;
  begin
    Result := -1;
    for i := 0 to form.ComponentCount-1 do begin
      if form.Components[i] is hlClass then begin
        Result := i;
        Exit;
      end;
    end; //for
  end;

  procedure PlaceNew;
  var
    i: integer;
    high: integer;
    comp: integer;
    xpos, ypos: integer;
    xstart: integer;

    procedure GetStartCoordinates;
    var
      compTop: integer;
      compLeft: integer;
      i: integer;
    begin
      xpos := -1;
      ypos := -1;
      for i := 0 to form.ComponentCount-1 do begin
        if form.Components[i] is TmwCustomHighlighterClass then begin
          compLeft := LongRec(form.Components[i].DesignInfo).Lo;
          compTop  := LongRec(form.Components[i].DesignInfo).Hi;
          if (xpos < 0) or (compLeft < xpos) then
            xpos := compLeft;
          if (ypos < 0) or (compTop < ypos) then
            ypos := compTop;
        end;
      end; //for
      if xpos < 0 then
        xpos := 8;
      if ypos < 0 then
        ypos := 8;
      xstart := xpos;
    end;

    procedure IncCoordinates;
    begin
      Inc(xpos,32);
      if (xpos+32) >= form.ClientWidth then begin
        xpos := xstart;
        Inc(ypos,32);
      end;
    end;

    function CoordinatesTaken: boolean;
    var
      compTop: integer;
      compLeft: integer;
      compRect: TRect;
      testRect: TRect;
      interRect: TRect;
      i: integer;
    begin
      Result := false;
      testRect := Rect(xpos,ypos,xpos+31,ypos+31);
      for i := 0 to form.ComponentCount-1 do begin
        if (form.Components[i] <> self) and (not (form.Components[i] is TControl)) then begin
          compLeft := LongRec(form.Components[i].DesignInfo).Lo;
          compTop  := LongRec(form.Components[i].DesignInfo).Hi;
          compRect := Rect(compLeft,compTop,compLeft+31,compTop+31);
          if IntersectRect(interRect,testRect,compRect) then begin
            Result := true;
            Exit;
          end;
        end;
      end; //for
    end;

    procedure GetFreeCoordinates;
    begin
      while CoordinatesTaken do
        IncCoordinates;
    end;

  begin
    GetStartCoordinates;
    // Iterate over TCheckListBox, not over GetPlaceableHighlighters to ensure
    // inserted highlighters to be sorted by name.
    // Iterate twice - delete highlighters in first pass (to make place), create
    // in second.
    for i := 0 to mwForm.clbHighlighters.Items.Count-1 do begin
      if not mwForm.clbHighlighters.Checked[i] then begin // unchecked - remove
        high := highlight.FindByName(mwForm.clbHighlighters.Items[i]);
        if high >= 0 then begin
          comp := FindHighlighterComp(highlight[high]);
          if comp >= 0 then
            form.Components[comp].Free;
        end;
      end;
    end; //for
    for i := 0 to mwForm.clbHighlighters.Items.Count-1 do begin
      if mwForm.clbHighlighters.Checked[i] then begin // checked - add
        high := highlight.FindByName(mwForm.clbHighlighters.Items[i]);
        if high >= 0 then begin
          if FindHighlighterComp(highlight[high]) < 0 then begin
            GetFreeCoordinates;
            dsgn.CreateComponent(highlight[high],AOwner,xpos,ypos,24,24);
            IncCoordinates;
          end;
        end;
      end;
    end; //for
  end;

begin
  inherited;
  if (csDesigning in ComponentState) and (AOwner is TCustomForm) then begin
    form := TCustomForm(AOwner);
    dsgn := form.Designer as {$IFDEF MWE_COMPILER_4_UP}IFormDesigner{$ELSE}TFormDesigner{$ENDIF};
    highlight := GetPlaceableHighlighters;
    if highlight.Count = 0 then
      Application.MessageBox('No highlighters found!','Highlighter Manager',MB_OK+MB_ICONEXCLAMATION)
    else begin
      mwForm := TmwHighlighterForm.Create(highlight);
      try
        CheckExisting;
        if mwForm.ShowModal = mrOK then
          PlaceNew;
      finally mwForm.Free; end;
    end;
  end;
  Abort;
end;

{ TmwHighlighterForm }

constructor TmwHighlighterForm.Create(highlighters: THighlighterList);
begin
  CreateNew(nil);
  Caption := 'Highlighter Manager';
  Width  := 410;
  Height := 243;
  Position := poScreenCenter;
  BorderStyle := bsDialog;

  Highlight := highlighters;
  
//object clbHighlighters: TCheckListBox
//  Left = 8
//  Top = 8
//  Width = 305
//  Height = 201
//  ItemHeight = 13
//  TabOrder = 0
//end

//object btnSelectAll: TButton
//  Left = 320
//  Top = 8
//  Width = 75
//  Height = 25
//  Caption = '&Select All'
//  TabOrder = 1
//end

//object btnDeselectAll: TButton
//  Left = 320
//  Top = 40
//  Width = 75
//  Height = 25
//  Caption = '&Deselect All'
//  TabOrder = 2
//end

//object btnOK: TButton
//  Left = 320
//  Top = 152
//  Width = 75
//  Height = 25
//  Caption = 'OK'
//  Default = True
//  ModalResult = 1
//  TabOrder = 3
//end

//object btnCancel: TButton
//  Left = 320
//  Top = 184
//  Width = 75
//  Height = 25
//  Caption = 'Cancel'
//  ModalResult = 2
//  TabOrder = 4
//end

  clbHighlighters := TCheckListBox.Create(Self);
  btnSelectAll := TButton.Create(Self);
  btnDeselectAll := TButton.Create(Self);
  btnOK := TButton.Create(Self);
  btnCancel := TButton.Create(Self);
  with clbHighlighters do
  begin
    Name := 'clbHighlighters';
    Parent := Self;
    Left := 8;
    Top := 8;
    Width := 305;
    Height := 201;
    ItemHeight := 13;
    Sorted := true;
    TabOrder := 0;
  end;
  with btnSelectAll do
  begin
    Name := 'btnSelectAll';
    Parent := Self;
    Left := 320;
    Top := 8;
    Width := 75;
    Height := 25;
    Caption := '&Select All';
    TabOrder := 1;
    OnClick := SelectAll;
  end;
  with btnDeselectAll do
  begin
    Name := 'btnDeselectAll';
    Parent := Self;
    Left := 320;
    Top := 40;
    Width := 75;
    Height := 25;
    Caption := '&Deselect All';
    TabOrder := 2;
    OnClick := DeselectAll;
  end;
  with btnOK do
  begin
    Name := 'btnOK';
    Parent := Self;
    Left := 320;
    Top := 152;
    Width := 75;
    Height := 25;
    Caption := 'OK';
    Default := True;
    ModalResult := 1;
    TabOrder := 3;
  end;
  with btnCancel do
  begin
    Name := 'btnCancel';
    Parent := Self;
    Left := 320;
    Top := 184;
    Width := 75;
    Height := 25;
    Caption := 'Cancel';
    ModalResult := 2;
    TabOrder := 4;
  end;
  LoadForm;
end;

procedure TmwHighlighterForm.DeselectAll(Sender: TObject);
var
  i: integer;
begin
  for i := 0 to clbHighlighters.Items.Count-1 do
    clbHighlighters.Checked[i] := false;
end;

procedure TmwHighlighterForm.LoadForm;
var
  i: integer;
begin
  clbHighlighters.Clear;
  for i := 0 to Highlight.Count-1 do begin
    clbHighlighters.Items.Add(Highlight[i].GetLanguageName); 
  end; //for
end;

procedure TmwHighlighterForm.SelectAll(Sender: TObject);
var
  i: integer;
begin
  for i := 0 to clbHighlighters.Items.Count-1 do
    clbHighlighters.Checked[i] := true;
end;

end.

